import {
	Directive, ComponentFactoryResolver, ComponentRef, ViewContainerRef,
	Input, Output, EventEmitter, OnInit, OnChanges, OnDestroy,
	Optional, Self, SkipSelf, Inject
} from '@angular/core';
import {
	Form, FormControl, AbstractControl,
	NgControl, ControlContainer, ControlValueAccessor,
	NG_VALIDATORS, Validator, Validators, ValidatorFn,
	NG_ASYNC_VALIDATORS, AsyncValidatorFn
} from '@angular/forms';


import { ffcArr } from './../components';
// const formFieldComponents = ffcArr.reduce((acc, componentClass) => ({ ...acc, [componentClass.name]: componentClass }), {}); // Works with JIT, but not AOT
const getFormFieldComponents = () => {
	return ffcArr.reduce((acc, componentClass) => {
		try {
			const componentName = new (componentClass as any)({}, {}).componentName; // Work around AOT name-mangling by explicitly storing the name as a class property
			if (!componentName) {
				throw new Error(`All components must have an explicit component name - set componentName property in component's class`);
			}
			return { ...acc, [componentName]: componentClass }
		} catch(e) {
			console.error(`DYNAFORM: ERROR COMPILING HASH OF FORM FIELD COMPONENTS\n${e}`);
			console.error(componentClass);
			return acc;
		}
	}, {});
}
const formFieldComponents = getFormFieldComponents();


interface IFFC {
	control: FormControl; // Remember, this can be an individual FormControl or a FormGroup
	meta: StringMap<any>;
	propagateChange?: Function;
	call?: EventEmitter<string>;
}
type FFCCustom = IFFC & ControlValueAccessor;

// Generate component name given type
const componentType = (type: string): string => type[0].toUpperCase() + type.slice(1) + 'Component';

@Directive({
	// tslint:disable-next-line:directive-selector
	selector: '[dynafield]'
})
export class DynafieldDirective extends NgControl implements OnInit, OnChanges, OnDestroy {

	@Input()
	meta: StringMap<any>;

	@Input()
	set control(fc: FormControl) {
		this._control = this._control || fc;
	}

	get control(): FormControl | null {
		return this._control;
	}

	// tslint:disable-next-line:no-output-rename
	@Output('ngModelChange')
	update = new EventEmitter();

	@Output()
	call: EventEmitter<string> = new EventEmitter<string>();

	component: ComponentRef<IFFC|FFCCustom>;
	_control;

	constructor(
		private resolver: ComponentFactoryResolver,
		private container: ViewContainerRef,
		@SkipSelf() private cc: ControlContainer,
		@Optional() @Self() @Inject(NG_VALIDATORS) private validators: Array<Validator | ValidatorFn>,
		@Optional() @Self() @Inject(NG_ASYNC_VALIDATORS) private asyncValidators: AsyncValidatorFn[]
	) {
		super();
	}

	ngOnInit() {
		const type = componentType(this.meta.type);
		if (!formFieldComponents[type]) {
			const validComponentTypes = Object.keys(formFieldComponents).join(', ');
			throw new Error(
				`Dynaform Dynafield Error: Invalid field type: ${type}.
				 Supported types: ${validComponentTypes}`
			);
		}
		try {
			let { control, meta } = this;
			const { name, class: cssClass, id: cssId, disabled } = meta;

			// Create the component
			const componentFactory = this.resolver.resolveComponentFactory<IFFC>(formFieldComponents[type]);
			this.component = this.container.createComponent(componentFactory);
			const instance = this.component.instance;

			// Support the recursive insertion of Dynaform components
			if (type === 'DynaformComponent') {
				if (meta.template) {
					(instance as any).template = meta.template;
				}
				meta = meta.meta;
			}

			// Set id and classes
			this.setCssId(cssId);
			this.setCssClasses(type, cssClass);

			// Check whether it's disabled, then set its FormControl and metadata
			if (disabled) {
				this.control.reset({ value: this.control.value, disabled: true });
			}
			instance.control = control;
			instance.meta = meta;

			// Listen for 'call' Output events and send them onwards
			if (instance.call) {
				instance.call.subscribe((fnId: string) => this.call.emit(fnId));
			}

			// Connect custom components
			if (instance.propagateChange) {
				// We're dealing with a custom form control which implements the ControlValueAccessor interface,
				// so we need to wire it up!
				this.name = name;
				this.valueAccessor = this.component.instance as FFCCustom;
				// Attach sync validators
				/*
				const ngValidators = this.component.injector.get(NG_VALIDATORS, null);
				if (ngValidators && ngValidators.some(x => x as any === this.component.instance)) {
					console.log('HERE');
					this.validators = [...(this.validators || []), ...(ngValidators as Array<Validator|ValidatorFn>)];
				}
				*/
				this._control = this.formGroupDirective.addControl(this as NgControl);
			}
		} catch (e) {
			console.error('ERROR INSTANTIATING DYNAFORM CHILD COMPONENT', type);
			console.log(e);
		}
	}

	ngOnChanges() {
		// We won't support mutating components (e.g. Text --> Select) at this stage,
		// but will support mutating an instantiated components metadata
		// As the component is created in ngOnInt this does nothing in the run before ngOnInit, but responds to later input changes
		if (this.component) {
			const { type, class: cssClass, id: cssId } = this.meta;
			this.setCssId(cssId);
			this.setCssClasses(type, cssClass);
			this.component.instance.meta = this.meta;
		}
	}

	ngOnDestroy(): void {
		if (this.formGroupDirective) {
			this.formGroupDirective.removeControl(this as NgControl);
		}
		if (this.component) {
			this.component.destroy();
		}
	}

	// ---------------------------------------
	// Override methods / getters in NgControl

	viewToModelUpdate(newValue: any): void {
		this.update.emit(newValue);
	}

	get path(): string[] {
		return [...this.cc.path, this.name] as string[];
	}

	get formGroupDirective(): Form | null {
		return this.cc ? this.cc.formDirective : null;
	}

	get validator(): ValidatorFn | null {
		return this.validators !== null ? Validators.compose(this.validators.map(this.normalizeValidator)) : null;
	}

	get asyncValidator(): AsyncValidatorFn {
		// TODO: Test composition of asyncValidators properly
		return this.asyncValidators !== null ? Validators.composeAsync(this.asyncValidators) : null;
	}

	normalizeValidator(validator: ValidatorFn | Validator): ValidatorFn {
		if ((validator as Validator).validate) {
			return (c: AbstractControl) => (validator as Validator).validate(c);
		} else {
			return validator as ValidatorFn;
		}
	}

	setCssId(cssId: string): void {
		const el = this.component.location.nativeElement;
		cssId ? el.setAttribute('id', cssId) : el.removeAttribute('id');
	}

	setCssClasses(type: string, cssClass: string|string[]): void {
		const classList = this.component.location.nativeElement.classList as DOMTokenList;
		const componentTypeClass = type.toLowerCase().replace('component', '');
		classList.add(componentTypeClass);
		if (typeof cssClass === 'string' || Array.isArray(cssClass)) {
			const classesToAdd = (Array.isArray(cssClass) ? cssClass : [...cssClass.split(/\s+/)]).filter(Boolean);
			const classesToRemove = [];
			const dontRemove = [ ...classesToAdd, componentTypeClass, 'ng-star-inserted' ];
			classList.forEach(c => classesToRemove.push(dontRemove.indexOf(c) < 0 ? c : null));
			classList.add(...classesToAdd);
			classList.remove(...classesToRemove);
		}
	}

}

